/**************************************************************************
*                                                                         *
*   KnightAndBishop                                                       *
*                                                                         *
*   Copyright (c) 2010 by Nedeljko Stefanovic, stenedjo@yahoo.com.        *
*                                                                         *
*   You are free to use, copy, modify and distribute it under the terms   *
*   of either version 2 or version 3 of the GNU General Public License    *
*   as published by the Free Software Foundation.                         *
*                                                                         *
**************************************************************************/



#include <iostream>
#include <fstream>
#include <cstdio>
#include <cctype>
#include <cstdlib>
#include <string>

using namespace std;


class Field {
    int rowNumber, columnNumber;
    int fieldCode;

public:
    Field() {
        fieldCode = rowNumber = columnNumber = 0;
    }

    Field(int row, int column) {
        rowNumber = row;
        columnNumber = column;
        fieldCode = (row << 3) | column;
    }

    Field(int code) {
        fieldCode = code;
        rowNumber = code >> 3;
        columnNumber = code & 7;
    }

    bool operator==(const Field &field) const {
        return fieldCode == field.fieldCode;
    }

    bool operator!=(const Field &field) const {
        return fieldCode != field.fieldCode;
    }

    int row() const {
        return rowNumber;
    }

    int column() const {
        return columnNumber;
    }

    void move(int dr, int dc) {
        rowNumber += dr;
        columnNumber += dc;
        fieldCode = (rowNumber << 3) | columnNumber;
    }

    void moveTo(int row, int column) {
        rowNumber = row;
        columnNumber = column;
        fieldCode = (row << 3) | column;
    }

    int code() const {
        return fieldCode;
    }

    void set(int code) {
        fieldCode = code;
        rowNumber = code >> 3;
        columnNumber = code & 7;
    }

    bool legal() const {
        return rowNumber == (rowNumber & 7) && columnNumber == (columnNumber & 7);
    }

    string toString() const {
        string result;

        result += (char)('a' + columnNumber);
        result += (char)('1' + rowNumber);

        return result;
    }
};

enum Side {
    BLACK,
    WHITE
};

class Position {
    friend void initialize();

public:
    Field black, white, knight, bishop;
    Side turn;

public:
    Position(int code) {
        bishop = Field(code & 63);
        code >>= 6;
        knight = Field(code & 63);
        code >>= 6;
        white = Field(code & 63);
        code >>= 6;
        black = Field(code & 63);
        turn = (Side) (code >> 6);
    }

    bool operator==(const Position &position) const {
        return black == position.black &&
            white == position.white &&
            knight == position.knight &&
            bishop == position.bishop &&
            turn == position.turn;
    }

    bool operator!=(const Position &position) const {
        return black != position.black ||
            white != position.white ||
            knight != position.knight ||
            bishop != position.bishop ||
            turn != position.turn;
    }

    int code() const {
        return (turn << 24) |
            (black.code() << 18) |
            (white.code() << 12) |
            (knight.code() << 6) |
            bishop.code();
    }

    bool check() const {
        if (abs(black.row() - white.row()) <= 1 && abs(black.column() - white.column()) <= 1) {
            return true;
        }

        if (abs((knight.row() - black.row())*(knight.column() - black.column())) == 2) {
            return true;
        }

        int dr = bishop.row() - black.row();
        int dc = bishop.column() - black.column();

        if (abs(dr) == abs(dc)) {
            int dr2 = white.row() - black.row();
            int dc2 = white.column() - black.column();

            if (abs(dr2) != abs(dc2) || (dr*dr2 <= 0) || (dc*dc2 <= 0) || (abs(dr2) >= abs(dr))) {
                dr2 = knight.row() - black.row();
                dc2 = knight.column() - black.column();

                if (abs(dr2) != abs(dc2) || (dr*dr2 <= 0) || (dc*dc2 <= 0) || (abs(dr2) >= abs(dr))) {
                    return true;
                }
            }
        }

        return false;
    }

    bool legal() const {
        if ((black.legal() == false) ||
            (white.legal() == false) ||
            (knight.legal() == false) ||
            (bishop.legal() == false)) {
            return false;
        }

        if (black == white ||
            knight == white ||
            bishop == white ||
            knight == black ||
            bishop == black ||
            knight == bishop) {
            return false;
        }

        if (turn == BLACK) {
            return abs(black.row() - white.row()) > 1 || abs(black.column() - white.column()) > 1;
        }

        return check() == false;
    }
};

struct Values {
    char *values;

    Values();
    ~Values() {
        delete [] values;
        values = 0;
    }
};

class Computation : public Position {
    mutable bool computed;
    static char* initialization();

protected:
    static Values val;
    mutable char positionValue;
    mutable bool stop;

    virtual void beforeLoop() const {
    }

    virtual void afterLoop() const {
    }

    virtual void iteration(const Position &position) const {
    }

private:

#define DO_ITERATION \
    if (current.legal()) { \
        iteration(current); \
        \
        if (stop) { \
            afterLoop(); \
            \
            return; \
        } \
    } \

    void allMoves() const {
        Position current(*this);

        if (current.turn == BLACK) {
            current.turn = WHITE;

            for (int dr = -1; dr <= 1; ++dr) {
                for (int dc = -1; dc <= 1; ++dc) {
                    if (dr == 0 && dc == 0) {
                        continue;
                    }

                    current.black.move(dr, dc);
                    DO_ITERATION
                    current.black = black;
                }
            }

            current.turn = BLACK;

            return;
        }

        current.turn = BLACK;

        for (int dr = -1; dr <= 1; ++dr) {
            for (int dc = -1; dc <= 1; ++dc) {
                if (dr == 0 && dc == 0) {
                    continue;
                }

                current.white.move(dr, dc);
                DO_ITERATION
                current.white = white;
            }
        }

        current.turn = BLACK;
        current.knight.move(-2, -1);
        DO_ITERATION
        current.knight.move(0, 2);
        DO_ITERATION
        current.knight.move(1, -3);
        DO_ITERATION
        current.knight.move(0, 4);
        DO_ITERATION
        current.knight.move(2, -4);
        DO_ITERATION
        current.knight.move(0, 4);
        DO_ITERATION
        current.knight.move(1, -3);
        DO_ITERATION
        current.knight.move(0, 2);
        DO_ITERATION
        current.knight = knight;

        for (int dr = -1; dr <= 1; dr += 2) {
            for (int dc = -1; dc <= 1; dc += 2) {
                for (int d = 1; d <= 7; ++d) {
                    current.bishop.move(dr*d, dc*d);

                    DO_ITERATION else {
                        current.bishop = bishop;

                        break;
                    }

                    current.bishop = bishop;
                }
            }
        }

        current.turn = WHITE;
    }

#undef DO_ITERATION

    void compute() const {
        stop = false;
        beforeLoop();
        allMoves();
        afterLoop();
        computed = true;
    }

public:

    Computation(int code) : Position(code) {
        computed = false;
    }

    char value() const {
        if (computed == false) {
            compute();
        }

        return positionValue;
    }
};

Values Computation::val;

class SimpleComputation : public Computation {
    mutable bool isCheckmate;
    mutable bool isStallmate;
    mutable bool isDraw;

protected:
    void beforeLoop() const {
        Position current(*this);

        isStallmate = true;
        isDraw = false;
        positionValue = -1;

        if (current.turn == WHITE) {
            return;
        }

        if (abs(current.black.row() - current.knight.row()) <= 1 &&
            abs(current.black.column() - current.knight.column()) <= 1) {
            current.black = current.knight;

            if (current.check() == false) {
                isDraw = true;
            }

            current.black = black;
        }

        if (isDraw == false &&
            abs(current.black.row() - current.bishop.row()) <= 1 &&
            abs(current.black.column() - current.bishop.column()) <= 1) {
            Field oldPosition = black;

            current.black = current.bishop;
            current.bishop.move(1, 0);

            if (current.check() == false) {
                isDraw = true;
            }

            current.bishop = bishop;
            current.black = black;
        }

        if (isDraw) {
            isStallmate = false;
            positionValue = 51;
            stop = true;
        }
    }

    void iteration(const Position &position) const {
        isStallmate = false;
        stop = true;
    }

    void afterLoop() const {
        if (isStallmate) {
            if (turn == BLACK && check()) {
                isCheckmate = true;
                isStallmate = false;
                isDraw = false;
                positionValue = 0;
            } else {
                isCheckmate = false;
                isDraw = true;
                positionValue = 51;
            }
        } else {
            isCheckmate = false;
        }
    }

public:
    SimpleComputation(int code) : Computation(code) {
    }
};

class WhiteWinComputation : public Computation {
    int iterationNumber;

protected:
    void beforeLoop() {
        positionValue = 0;
    }

    void iteration(const Position &position) const {
        Position current(position);

        current.turn = WHITE;

        int code = current.code();

        if (val.values[code] == -1) {
            positionValue = 1;
            val.values[code] = iterationNumber;
        }
    }

public:
    WhiteWinComputation(int code, int iterationNumber) : Computation(code) {
        this->iterationNumber = iterationNumber;
    }
};

class ComplexComputation : public Computation {
    int iterationNumber;
    mutable bool undefinedFound;

protected:
    void beforeLoop() const {
        undefinedFound = false;
        positionValue = -1;
    }

    void afterLoop() const {
        if (undefinedFound == false && positionValue == -1) {
            positionValue = turn == WHITE ? 51 : iterationNumber + 1;
        }
    }

    void iteration(const Position &position) const {
        int code = position.code();
        int value = val.values[code];

        if (turn == WHITE) {
            if (value == iterationNumber) {
                positionValue = iterationNumber;
                stop = true;
            } else if (value == -1) {
                undefinedFound = true;
            }

            return;
        }

        if (value == 51) {
            positionValue = 51;
            stop = true;
        } else if (value == -1) {
            undefinedFound = true;
        }
    }

public:
    ComplexComputation(int code, int iterationNumber) : Computation(code) {
        this->iterationNumber = iterationNumber;
    }
};

class BestMove : public Computation {
    mutable string bestMove;
    mutable Position bestPosition;
    mutable int bestCode;
    mutable int bestValue;

protected:
    void beforeLoop() const {
        if (turn == BLACK) {
            bestValue = -2;
        } else {
            bestValue = 52;
        }
    }

    void afterLoop() const {
        Position position(bestCode);

        bestPosition.black = position.black;
        bestPosition.white = position.white;
        bestPosition.knight = position.knight;
        bestPosition.bishop = position.bishop;
        bestPosition.turn = position.turn;

        if (bestPosition.black != black) {
            bestMove = "K";
            bestMove += bestPosition.black.toString();
        } else if (bestPosition.white != white) {
            bestMove = "K";
            bestMove += bestPosition.white.toString();
        } else if (bestPosition.knight != knight) {
            bestMove = "N";
            bestMove += bestPosition.knight.toString();
        } else if (bestPosition.bishop != bishop) {
            bestMove = "B";
            bestMove += bestPosition.bishop.toString();
        } else {
            throw("Checkmate");
        }

        if (bestPosition.turn == BLACK) {
            if (bestPosition.check()) {
                bestMove += "+";

                if (val.values[code()] == 0) {
                    bestMove += "+";
                }
            }
        }
    }

    void iteration(const Position &position) const {
        int code = position.code();
        int value = val.values[code];

        if (turn == BLACK) {
            if (value > bestValue) {
                bestValue = value;
                bestCode = code;
            }
        } else if (turn == WHITE) {
            if (value < bestValue) {
                bestValue = value;
                bestCode = code;
            }
        }
    }

public:
    BestMove(int code) : Computation(code), bestPosition(0) {
    }

    string toString() const {
        return bestMove;
    }

    Position newPosition() const {
        return bestPosition;
    }
};

Values::Values() {
    cout << "KnightAndBishop\n" << endl;
    cout << "Copyright (c) 2010 by Nedeljko Stefanovic, stenedjo@yahoo.com.\n" << endl;
    cout << "You are free to use, copy, modify and distribute it under the terms" << endl;
    cout << "of either version 2 or version 3 of the GNU General Public License" << endl;
    cout << "as published by the Free Software Foundation.\n\n" << endl;

    const char fileName[] = "KnightAndBishop.dat";
    ifstream in(fileName, ios_base::binary);

    values = new char[1 << 25];

    if (in) {
        cout << "Reading data" << endl;
        in.read(values, 1 << 25);

        if (in.good()) {
            cout << "Data readed" << endl;

            return;
        }

        cout << "Reading error" << endl;
    }

    cout << "Data not present. Initialization\n" << endl;

    for (int code = 0; code < (1 << 25); ++code) {
        SimpleComputation position(code);

        if (position.legal()) {
            values[code] = position.value();
        } else {
            values[code] = 127;
        }
    }

    for (int iterationNumber = 0; iterationNumber <= 50; ++iterationNumber) {
        cout << "\rStage " << iterationNumber << " white" << flush;

        bool changed = false;

        for (int code = 0; code < (1 << 25); ++code) {
            if (values[code] == -1) {
                ComplexComputation position(code, iterationNumber);

                if (position.turn == WHITE) {
                    values[code] = position.value();

                    if (values[code] != -1) {
                        changed = true;
                    }
                }
            }
        }

        if (changed == false) {
            break;
        }

        changed = false;
        cout << "\rStage " << iterationNumber << " black" << flush;

        for (int code = 0; code < (1 << 25); ++code) {
            if (values[code] == -1) {
                ComplexComputation position(code, iterationNumber);

                if (position.turn == BLACK) {
                    values[code] = position.value();

                    if (values[code] != -1) {
                        changed = true;
                    }
                }
            }
        }

        if (changed == false) {
            break;
        }
    }

    cout << "\n\nSaving data" << endl;

    ofstream out(fileName, ios_base::binary);

    if (out) {
        out.write(values, 1 << 25);

        if (out.good()) {
            cout << "Data saved" << endl;

            return;
        }

        cout << "Writing error" << endl;
        remove(fileName);
        delete [] values;
        values = 0;
    } else {
        cout << "Writing impossible" << endl;
        delete [] values;
        values = 0;
    }
}

istream& operator>>(istream &in, Field &field) {
    string position;
    int row = -1, column = -1;

    do {
        in >> position;

        if (position.size() == 2) {
            position[0] = tolower(position[0]);
            row = position[1] - '1';
            column = position[0] - 'a';
        }
    } while (row < 0 || column < 0 || row > 7 || column > 7);

    field.moveTo(row, column);

    return in;
}

class Inspector : public Computation {
protected:
    void beforeLoop() const {
        positionValue = val.values[code()];
        stop = true;
    }

public:
    Inspector(int code) : Computation(code) {
    }

    bool isOK() const {
        return val.values != 0;
    }
};

class Legality : public Computation {
    Position targetPosition;

    void beforeLoop() const {
        positionValue = 0;
    }

    void iteration(const Position &position) const {
        if (position == targetPosition) {
            positionValue = 1;
            stop = true;
        }
    }

public:
    Legality(int from, int to) : Computation(from), targetPosition(to) {
    }
};

int main(int argc, char *argv[]) {
    if (Inspector(0).isOK() == false) {
        cout << "Error! Initialization failed." << endl;

        return EXIT_FAILURE;
    }

    Position position(0);

    cout << "\n\nEnter the positions of the pieces\n" << endl;
    cout << "White king " << flush;
    cin >> position.white;
    cout << "White knight " << flush;
    cin >> position.knight;
    cout << "White bishop " << flush;
    cin >> position.bishop;
    cout << "Black king " << flush;
    cin >> position.black;

    string turn;
    bool turnReaded = false;

    cout << "Turn (black or white) " << flush;

    do {
        cin >> turn;

        if (turn.length() > 0) {
            if (turn[0] == 'w' || turn[0] == 'W') {
                position.turn = WHITE;
                turnReaded = true;
            } else if (turn[0] == 'b' || turn[0] == 'B') {
                position.turn = BLACK;
                turnReaded = true;
            }
        }
    } while (turnReaded == false);

    if (position.legal() == false) {
        cout << "Illegal position" << endl;

        return EXIT_SUCCESS;
    }

    cout << endl;

    int counter = 0;

    if (position.turn == BLACK) {
        if (Inspector(position.code()).value() == 0) {
            cout << "Checkmate position" << endl;

            return EXIT_SUCCESS;
        }

        cout << "1. ..." << flush;
        counter = 1;
    }

    try {
        while (true) {
            string move;

            if (position.turn == WHITE) {
                cout << ++counter << ". ";

                BestMove pos(position.code());

                pos.value();
                move = pos.toString();
                cout << move << flush;
                position = pos.newPosition();

                if (move.size() >= 2 && move.substr(move.size() - 2) == "++") {
                    break;
                }
            } else {
                cout << " K" << flush;

                while (true) {
                    Field newField;

                    cin >> newField;

                    Position newPosition(position);

                    newPosition.black = newField;
                    newPosition.turn = WHITE;

                    if (Legality(position.code(), newPosition.code()).value() == 1) {
                        position = newPosition;

                        break;
                    } else {
                        cout << "Illegal move" << endl;
                    }
                }
            }
        }
    } catch (const char *message) {
        cout << message << endl;
    }


    return EXIT_SUCCESS;
}
